/*
 -------------------------------------------------------------------------
 Copyright (c) 2002-2006, XIMETA, Inc., FREMONT, CA, USA.
 All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary 
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright 
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products 
      built using this software without specific written permission. 
      
 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.
 
 DISCLAIMER

 This software is provided 'as is' with no explcit or implied warranties
 in respect of any properties, including, but not limited to, correctness 
 and fitness for purpose.
 -------------------------------------------------------------------------
*/

/*
 * Data stuctures and Functions to provide the NDAS block device in linux
 * 
 */
 
#ifndef _LINUX_BLOCK_H_
#define _LINUX_BLOCK_H_

#include <sal/mem.h>
#include <ndasuser/info.h> // ndas_slot_info

#ifdef DEBUG
#include <sal/debug.h>
#if !defined(DEBUG_LEVEL_BLOCK) 
    #if defined(DEBUG_LEVEL)
    #define DEBUG_LEVEL_BLOCK DEBUG_LEVEL
    #else
    #define DEBUG_LEVEL_BLOCK 1
    #endif
#endif
#define dbgl_blk(l,x...) do { \
    if ( l <= DEBUG_LEVEL_BLOCK ) {\
        sal_debug_print("BL|%d|%s|",l,__FUNCTION__);\
        sal_debug_println(x); \
    } \
} while(0) 
#else
#define dbgl_blk(l,fmt...) do {} while(0)
#endif


#if LINUX_VERSION_25_ABOVE // 2.5.x or above
#define DRIVER_MASK        0xf0
#define PARTN_BITS        4    
#else
#define DRIVER_MASK        0xf8
#define PARTN_BITS        3    
#endif
#define NR_PARTITION    (1U << PARTN_BITS)    
#define PARTN_MASK        (NR_PARTITION -1)

#define NR_UNIT            1
#define UNITN_BITS        0
#define UNITN_MASK        0x00

#define DEFAULT_ND_BLKSIZE    4096    
#define DEFAULT_ND_BLKSIZE_BIT    12  
#define DEFAULT_ND_DEVSIZE    4
#define DEFAULT_ND_HARDSECTSIZE 512
#define DEFAULT_ND_MAX_SECTORS    128

#define ND_MAX_REQUEST         1

#define ND_MAJOR    60

#if LINUX_VERSION_25_ABOVE // 2.5.x or above
#include <linux/blkdev.h>
#else // 2.4.x or below
#define MAJOR_NR ND_MAJOR
#include <linux/major.h>
#if( MAJOR_NR == ND_MAJOR)
#undef    DEVICE_NAME
#define    DEVICE_NAME    "ndasdevice"
#undef    DEVICE_REQUEST
#define DEVICE_REQUEST    nblk_request_proc
#define DEVICE_NR(device) (MINOR(device) >> PARTN_BITS)
#undef    DEVICE_ON
#define DEVICE_ON(device)
#undef    DEVICE_OFF
#define DEVICE_OFF(device)
#endif
#include <linux/blk.h>

#endif // LINUX_VERSION

#if LINUX_VERSION_25_ABOVE

#define SLOT_I(_inode_) SLOT((_inode_)->i_bdev->bd_disk)
#define SLOT_R(_request_) SLOT((_request_)->rq_disk)
#define SLOT(_gendisk_) \
({ \
    struct gendisk *__g__ = _gendisk_;\
    (((__g__->first_minor) >> PARTN_BITS) + NDAS_FIRST_SLOT_NR);\
})

#else
#define SLOT_I(inode) SLOT((inode)->i_rdev)
#define SLOT_R(_request_) SLOT((_request_)->rq_dev)
#define SLOT(dev)            (DEVICE_NR(dev) + NDAS_FIRST_SLOT_NR)
#define PARTITION(dev)            ((dev) & PARTN_MASK) 
#define TOMINOR(slot, part)        (((slot - NDAS_FIRST_SLOT_NR) << PARTN_BITS) | (part))
#endif

/* 
 * Split the sectors into the given chunks for raid 0/4
 *
 * eg: 4 chunks(disks) from 128 sectors
 *    0-> 0   1->32   2->64   3->96
 *    4-> 1   5->33   6->65   7->97
 *   ...
 *  120->30 121->62 122->94 123->126
 *  124->31 125->63 126->95 127->127
 */
#define SPLIT_IDX(i, size, splits) \
    ( ((i) % (splits)) * ( (size) / (splits) ) + ((i) / (splits)))
    
#define SPLIT_IDX_SLOT(i, sz, slot) \
    SPLIT_IDX(i, \
        sz, \
        NDAS_GET_SLOT_DEV(slot)->info.io_splits)

/*
 * Get the byte size of the struct nbio with the given number of blocks
 */
#define NBIO_SIZE(nr_blocks) ( \
    sizeof(struct nbio) + \
    (nr_blocks) * sizeof(struct sal_mem_block))
/*
 * Decide if the nbio blocks and the given data address is adjacent to merge it
 */    
#define NBIO_ADJACENT(lio, data) ( \
    (lio)->nr_blocks > 0 && \
    (lio)->blocks[(lio)->nr_blocks - 1].ptr + \
    (lio)->blocks[(lio)->nr_blocks - 1].len == (data) )

#define NDAS_FLAG_QUEUE_SUSPENDED  (1<<0)


/*
 * The data structure to maintain the information about the slot
 */
struct ndas_slot {
    ndas_slot_info info;
    ndas_device_info dinfo;
    int            slot;
    int         enabled;     /* 1 if enabled */
    spinlock_t lock;
    struct semaphore mutex;
    ndas_error_t         enable_result;
#ifdef NDAS_HOTPLUG
    struct device dev;
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)) // 2.5.x or above
    struct gendisk *disk;
    /* 
      pointer to disk->queue, but will be deallocated on unregistration
     */
    struct request_queue *queue;

    unsigned long        queue_flags;   /* NDAS_FLAG_xxx */
    
#else
    unsigned int        flags[NR_PARTITION];
    int            ref[NR_PARTITION];
#endif


#ifdef NDAS_DEVFS
//	devfs_handle_t devfs_handle;
#endif

    struct proc_dir_entry *proc_dir;
    struct proc_dir_entry *proc_info;
    struct proc_dir_entry *proc_info2;
    struct proc_dir_entry *proc_unit;
    struct proc_dir_entry *proc_timeout;
    struct proc_dir_entry *proc_enabled;
    struct proc_dir_entry *proc_ndas_serial;
    struct list_head proc_node;

};

/* 
 * Structure to gather the scattered memory blocks.
 * NDAS driver is platform neutral driver, 
 * so we use the platform neutral data structure - sal_mem_block.
 */
struct nbio {
    struct request             *req;
    struct ndas_slot        *dev;
    int                     nr_blocks;
    size_t                     nr_sectors;
    struct sal_mem_block     blocks[0];
};
/*
 * Some old kernel such as 2.4.2 doesn't have max, min macros
 */
#define NB_MAX(x,y) ((x)>(y)?(x):(y))

#if LINUX_VERSION_25_ABOVE // 2.5.x or above
#define ND_BLK_MAX_REQ_SEGMENT 64        // Just arbitary number
#define MAX_NBIO_SIZE NB_MAX(NBIO_SIZE(ndas_io_unit * 2),NBIO_SIZE(ND_BLK_MAX_REQ_SEGMENT))
#else
#define MAX_NBIO_SIZE NB_MAX(NBIO_SIZE(ndas_io_unit * 2),NBIO_SIZE(MAX_SEGMENTS))
#endif

#if LINUX_VERSION_25_ABOVE // 2.5.x or above
/** 
 * For 2.6, we decided not to allocate gendisk for the whole NetDisk device in the insmod-ed time.
 * So the /dev/ndas is not available to ioctl.
 * We should assign a fake disk to /dev/ndas to allow the ndasadmin to call ioctl on /dev/ndas.
 * TODO: remove ioctl, and use procfs
 */
extern struct _special_disk {
    struct gendisk *gd;
    xbool mounted;
} special;
#else
extern struct hd_struct  ndas_hd[256];
extern int            ndas_sizes[256];
extern int            ndas_blksizes[256];
extern int            ndas_hardsectsizes[256];
extern int            ndas_max_sectors[256];
#endif

/*
 * Cached memory pool to be used for nbio 
 */
extern KMEM_CACHE *nblk_kmem_cache;

inline static void* nblk_mem_alloc(void) 
{
    return kmem_cache_alloc(nblk_kmem_cache, GFP_ATOMIC);
}
inline static void nblk_mem_free(void *object)
{
    kmem_cache_free(nblk_kmem_cache, object);
}

NDAS_CALL
void nblk_end_io(int slot, ndas_error_t err, struct nbio* nio);

ndas_error_t nblk_handle_io(struct request *req);

#endif // _LINUX_BLOCK_H_
